# -*- coding: utf-8 -*-
"""
Created on Thu Sep  5 13:05:12 2019

@author: icruicks
"""

import pandas as pd, GraphLearning as gl, networkx as nx, numpy as np
from scipy.sparse import csr_matrix, diags
from scipy.sparse.linalg import eigs
import os, warnings
from matplotlib import pyplot as plt

from sklearn.metrics import adjusted_rand_score, adjusted_mutual_info_score, confusion_matrix, silhouette_score
from sklearn.decomposition import LatentDirichletAllocation
from sklearn.neighbors import kneighbors_graph
from sklearn.cluster import SpectralClustering, AgglomerativeClustering, KMeans
from sklearn.cluster import SpectralCoclustering
from sknetwork.clustering import BiLouvain, Louvain, modularity
from Leiden_Clustering import Leiden_Unimodal
from sklearn.preprocessing import normalize

def create_twitter_network(users, edgelist):
    n = len(users)
    network = np.zeros([n,n])

    for i in range(n):
        pers_i = users[i]
        res_1 = edgelist[np.where(edgelist[:,0] ==pers_i)]
        for j in range(n):
            pers_j = users[j]
            res_2 = res_1[np.where(res_1[:,1] ==pers_j)]
            try:
                network[i,j] = res_2[:,2]
            except:
                network[i,j] = 0

    return csr_matrix(network)

def create_twitter_attribute(users, content):
    n = len(users)
    attributes = np.unique(content.iloc[:,0])
    m = len(attributes)
    network = np.zeros([n,m])
    for i in range(n):
        pers_i = users[i]
        res_1 = content[np.where(content.iloc[:,1] ==pers_i)]
        for j in range(m):
            attr_j = attributes[j]
            res_2 = res_1[np.where(res_1.iloc[:,0] == attr_j)]
            try:
                network[i,j] = res_2.iloc[:,2]
            except:
                network[i,j] = 0
    return csr_matrix(network)


def print_results(y_preds, y, model_name):
    aris = [adjusted_rand_score(y, y_pred) for y_pred in y_preds]
    amis = [adjusted_mutual_info_score(y, y_pred) for y_pred in y_preds]
    print("\n")
    print(model_name+" Results:")
    print("Adjusted Rand Score: {} +/- {}".format(np.mean(aris), np.std(aris)))
    print("Adjusted Mutual Information: {} +/- {}".format(np.mean(amis), np.std(amis)))
    # print("Confusion Matrix of best ARI result: \n {}".format(confusion_matrix(y, y_preds[np.argmax(aris)])))
    print("Avg. Number of clusters found: {}".format(np.mean([len(np.unique(y_pred)) for y_pred in y_preds])))

def cluster_comparison(y_pred, modal_clusters, modes):
    for i in range(len(modes)):
        print(modes[i])
        print("ARI: {}".format(adjusted_rand_score(modal_clusters[i], y_pred)))
        print("AMI: {}".format(adjusted_mutual_info_score(modal_clusters[i], y_pred)))

dataset = 'aucs'
num_iterations = 20
warnings.filterwarnings('ignore')

''' 
Read in Data
'''

if dataset =='cora':
    content = pd.read_csv(os.path.join('Data','Cora','cora_content.csv'), header=0, index_col=0)
    links = pd.read_csv(os.path.join('Data','Cora','cora_network_links.csv'), header=0, index_col=None)
    network = nx.from_pandas_edgelist(links, source='Source', target='Target', create_using=nx.DiGraph())
    df = nx.to_pandas_adjacency(network)
    content = content.reindex(labels=df.index, axis='index')
    adjacency = df.values
    X = content.iloc[:,:-1]
    y = content.iloc[:,-1].astype('category').cat.codes
    nearest_neighbors = 40
elif dataset=='citeseer':
    content = pd.read_csv(os.path.join('Data','CiteSeer','citeseer_content.csv'), header=0, index_col=0, dtype={0: object})
    links = pd.read_csv(os.path.join('Data','CiteSeer','citeseer_network_links.csv'), header=0, index_col=None, dtype=object)
    discrepancies = list(np.setdiff1d(links['Source'], content.index)) + list(np.setdiff1d(links['Target'], content.index))
    links = links[~links['Source'].isin(discrepancies)]
    links = links[~links['Target'].isin(discrepancies)]
    content = content[~content.index.isin(discrepancies)]
    network = nx.from_pandas_edgelist(links, source='Source', target='Target', create_using=nx.DiGraph())
    df = nx.to_pandas_adjacency(network)
    content = content.reindex(labels=df.index, axis='index')
    adjacency = df.values
    X = content.iloc[:,:-1]
    y = content.iloc[:,-1].astype('category').cat.codes
    nearest_neighbors = 40
elif dataset =='flickr':
    content = pd.read_csv(os.path.join('Data','Flickr','attributes.csv'), header=None)
    adjacency = pd.read_csv(os.path.join('Data','Flickr','network.csv'), header= None).to_numpy()
    y = pd.read_csv(os.path.join('Data','Flickr','labels.csv'), header = None)[0]
    X = content.to_numpy()
    nearest_neighbors = 60
elif dataset =='blogcatalog':
    content = pd.read_csv(os.path.join('Data','BlogCatalog','attributes.csv'), header=None)
    adjacency = pd.read_csv(os.path.join('Data','BlogCatalog','network.csv'), header= None).to_numpy()
    y = pd.read_csv(os.path.join('Data','BlogCatalog','labels.csv'), header = None)[0]
    X = content.to_numpy()
    nearest_neighbors = 60
elif dataset =='3sources':
    from scipy.io import mmread
    from functools import reduce
    bbc_content = csr_matrix(mmread(os.path.join('Data','3Sources','3sources_bbc.mtx'))).T
    guardian_content = csr_matrix(mmread(os.path.join('Data','3Sources','3sources_guardian.mtx'))).T
    reuters_content = csr_matrix(mmread(os.path.join('Data','3Sources','3sources_reuters.mtx'))).T
    labels = pd.read_csv(os.path.join('Data','3Sources','labels.csv'), header = None)[1].astype('category').cat.codes
    reuters_idx = np.ravel(pd.read_csv(os.path.join("Data","3sources","3sources_reuters.docs"), header=None).values)-1
    bbc_idx = np.ravel(pd.read_csv(os.path.join("Data","3sources","3sources_bbc.docs"), header=None).values)-1
    guardian_idx = np.ravel(pd.read_csv(os.path.join("Data","3sources","3sources_guardian.docs"), header=None).values)-1
    master_idx = reduce(np.intersect1d, (bbc_idx, guardian_idx, reuters_idx))
    bbc_X = bbc_content[np.where(np.isin(bbc_idx, master_idx))]
    reuters_X = reuters_content[np.where(np.isin(reuters_idx, master_idx))]
    guardian_X = guardian_content[np.where(np.isin(guardian_idx, master_idx))]
    y = labels[master_idx]
    nearest_neighbors = 20
elif dataset =='wiki':
    content = np.loadtxt(os.path.join('Data','Wiki','tfidf.txt'))
    links = np.loadtxt(os.path.join('Data','Wiki','graph.txt'))
    labels = np.loadtxt(os.path.join('Data','Wiki','group.txt'))
    X = pd.crosstab(content[:,0], content[:,1], values=content[:,2], aggfunc=sum)
    X = X.reindex(index=labels[:,0])
    X = np.nan_to_num(X.values)
    adjacency = pd.crosstab(links[:,0], links[:,1])
    adjacency = adjacency.reindex(index=labels[:,0], columns=labels[:,0])
    adjacency = csr_matrix(np.nan_to_num(adjacency.values))
    y= labels[:,1]
    nearest_neighbors = int(np.ceil(np.sqrt(labels.shape[0])))
elif dataset in ['football', 'olympics', 'politicsie', 'politicsuk', 'rugby']:
    user_ids = np.loadtxt(os.path.join("Data","Twitter",dataset,dataset+".ids"))
    content = np.loadtxt(os.path.join("Data","Twitter",dataset,dataset+"-follows.mtx"))[1:,:]
    follows_network = create_twitter_network(user_ids, content)
    content = np.loadtxt(os.path.join("Data","Twitter",dataset,dataset+"-mentions.mtx"))[1:,:]
    mentions_network = create_twitter_network(user_ids, content)
    content = np.loadtxt(os.path.join("Data","Twitter",dataset,dataset+"-retweets.mtx"))[1:,:]
    retweets_network = create_twitter_network(user_ids, content)
    content = np.loadtxt(os.path.join("Data","Twitter",dataset,dataset+"-lists500.mtx"))[1:,:]
    list_content = create_twitter_attribute(user_ids, content)
    content = np.loadtxt(os.path.join("Data","Twitter",dataset,dataset+"-tweets500.mtx"))[1:,:]
    text_content = create_twitter_attribute(user_ids, content)
    y = pd.read_csv(os.path.join("Data","Twitter",dataset,"labels.csv"), header = None)[1].astype('category').cat.codes
    nearest_neighbors = int(np.ceil(np.sqrt(user_ids.shape[0])))
elif dataset == 'fftwyt':
    links = pd.read_csv(os.path.join('Data','ff-tw-yt','raw_data.csv'), header= None)
    links_ff = links.loc[links[2] == 'ff']
    links_tw = links.loc[links[2] == 'tw']
    links_yt = links.loc[links[2] == 'yt']
    user_ids = pd.unique(links[[0, 1]].values.ravel('K'))
    ff_network = csr_matrix(pd.crosstab(links_ff[0], links_ff[1]).reindex(index=user_ids, columns=user_ids, fill_value=0).values)
    tw_network = csr_matrix(pd.crosstab(links_tw[0], links_tw[1]).reindex(index=user_ids, columns=user_ids, fill_value=0).values)
    yt_network = csr_matrix(pd.crosstab(links_yt[0], links_yt[1]).reindex(index=user_ids, columns=user_ids, fill_value=0).values)
elif dataset =='aucs':
    links = pd.read_csv(os.path.join('Data','AUCS','links.csv'), header= None)
    labels = pd.read_csv(os.path.join('Data','AUCS','attributes.csv'), header= None).fillna("none")
    user_ids= labels[0].values
    y= labels[1].astype('category').cat.codes
    links_lunch = links.loc[links[2] == 'lunch']
    lunch_network = csr_matrix(pd.crosstab(links_lunch[0], links_lunch[1]).reindex(index=user_ids, columns=user_ids, fill_value=0).values)
    links_facebook = links.loc[links[2] == 'facebook']
    facebook_network = csr_matrix(pd.crosstab(links_facebook[0], links_facebook[1]).reindex(index=user_ids, columns=user_ids, fill_value=0).values)
    links_coauthor = links.loc[links[2] == 'coauthor']
    coauthor_network = csr_matrix(pd.crosstab(links_coauthor[0], links_coauthor[1]).reindex(index=user_ids, columns=user_ids, fill_value=0).values)
    links_work = links.loc[links[2] == 'work']
    work_network = csr_matrix(pd.crosstab(links_work[0], links_work[1]).reindex(index=user_ids, columns=user_ids, fill_value=0).values)
    links_leisure = links.loc[links[2] == 'leisure']
    leisure_network = csr_matrix(pd.crosstab(links_leisure[0], links_leisure[1]).reindex(index=user_ids, columns=user_ids, fill_value=0).values)
else:
    raise ValueError(dataset+" not found")


cluster_graphs=[]
kNNs =[]
ba_matrices =[]
clusters =[]
    
''' 
Create the base local graphs and clusters 
'''
if dataset in ['cora', 'citeseer', 'blogcatalog', 'flickr', 'wiki']:
    # cluster_iterations = [Louvain(random_state=int(rand_state), shuffle_nodes=True) for rand_state in np.random.randint(1,high=100000,size=10)]
    # parts = [clstr.fit(adjacency).labels_ for clstr in cluster_iterations]
    cluster_iterations = [Leiden_Unimodal(resolution=i) for i in np.linspace(0.1,2.0,10)]
    parts = [clstr.fit_transform(adjacency) for clstr in cluster_iterations]
    modularities = [modularity(adjacency, part) for part in parts]
    # print("Modularity of network clustering: {}".format(np.max(modularities)))
    defacto_network_subgroups = parts[np.argmax(modularities)]
    clusters.append(defacto_network_subgroups)
    ba_matrix = np.zeros((defacto_network_subgroups.size, defacto_network_subgroups.max()+1))
    ba_matrix[np.arange(defacto_network_subgroups.size),defacto_network_subgroups] = 1
    ba_matrices.append(ba_matrix)
    co_matrix = np.matmul(ba_matrix, ba_matrix.T)
    #np.fill_diagonal(co_matrix, 0)
    cluster_graphs.append(csr_matrix(co_matrix))
    # kNN = np.maximum(adjacency, adjacency.T)
    # kNNs.append(csr_matrix(kNN))
    kNNs.append(csr_matrix(adjacency))
    
    
    kNN = kneighbors_graph(X, metric='cosine', mode='connectivity', n_neighbors=nearest_neighbors)
    kNN = kNN.minimum(kNN.T)
    kNNs.append(kNN)
    if dataset in ['cora', 'citeseer', 'blogcatalog', 'wiki']:
        # clstr = SpectralClustering(n_clusters=len(np.unique(y)), affinity='rbf', n_jobs=-1)
        # clstr = AgglomerativeClustering(n_clusters=len(np.unique(y)))
        # clstr = KMeans(n_clusters=len(np.unique(y)), random_state=1234)
        # clstr = OPTICS(metric='cosine')
        # content_subgroups = clstr.fit_predict(X)
        lda = LatentDirichletAllocation(n_components=len(np.unique(y)), learning_decay= 0.5, random_state=1234)
        lda_probs = lda.fit_transform(X)
        _,content_subgroups = np.unique(np.argmax(lda_probs, axis=1),return_inverse=True)
    elif dataset in ['flickr']:
        attrib_cluster_iterations = [Leiden_Unimodal(resolution=i) for i in np.linspace(0.1,2.0,10)]
        attrib_parts = [clstr.fit_transform(kNN) for clstr in cluster_iterations]
        attrib_modularities = [modularity(kNN, part) for part in attrib_parts]
        content_subgroups = attrib_parts[np.argmax(attrib_modularities)]
    # print("Silhouette of attributes clustering: {}".format(silhouette_score(X, content_subgroups)))
    clusters.append(content_subgroups)
    ba_matrix = np.zeros((content_subgroups.size, content_subgroups.max()+1))
    ba_matrix[np.arange(content_subgroups.size),content_subgroups] = 1
    ba_matrices.append(ba_matrix)
    co_matrix = np.matmul(ba_matrix, ba_matrix.T)
    #np.fill_diagonal(co_matrix, 0)
    cluster_graphs.append(csr_matrix(co_matrix))
elif dataset == '3sources':
    for X in [bbc_X, reuters_X, guardian_X]:
        clstr = AgglomerativeClustering(n_clusters=len(np.unique(y)))
        content_subgroups = clstr.fit_predict(X.toarray())
        # print("Silhouette of text clustering: {}".format(silhouette_score(X, content_subgroups)))
        clusters.append(content_subgroups)
        ba_matrix = np.zeros((content_subgroups.size, content_subgroups.max()+1))
        ba_matrix[np.arange(content_subgroups.size),content_subgroups] = 1
        ba_matrices.append(ba_matrix)
        co_matrix = np.matmul(ba_matrix, ba_matrix.T)
        cluster_graphs.append(csr_matrix(co_matrix))
        kNN = kneighbors_graph(X, metric='cosine', mode='connectivity', n_neighbors=nearest_neighbors)
        kNN = kNN.minimum(kNN.T)
        kNNs.append(kNN)
elif dataset in ['football', 'olympics', 'politicsie', 'politicsuk', 'rugby']:
    for X in [list_content, text_content]:
        lda = LatentDirichletAllocation(n_components=len(np.unique(y)), random_state=1234)
        lda_probs = lda.fit_transform(X)
        _,content_subgroups = np.unique(np.argmax(lda_probs, axis=1),return_inverse=True)
        # print("Silhouette of attributes clustering: {}".format(silhouette_score(X, content_subgroups)))
        clusters.append(content_subgroups)
        ba_matrix = np.zeros((content_subgroups.size, content_subgroups.max()+1))
        ba_matrix[np.arange(content_subgroups.size),content_subgroups] = 1
        ba_matrices.append(ba_matrix)
        co_matrix = np.matmul(ba_matrix, ba_matrix.T)
        cluster_graphs.append(csr_matrix(co_matrix))
        kNN = kneighbors_graph(X, metric='cosine', mode='connectivity', n_neighbors=nearest_neighbors)
        kNN = kNN.minimum(kNN.T)
        kNNs.append(kNN)
    for adjacency in [follows_network, mentions_network, retweets_network]:
        # cluster_iterations = [Louvain(random_state=int(rand_state), shuffle_nodes=True) for rand_state in np.random.randint(1,high=100000,size=5)]
        # parts = [clstr.fit(adjacency).labels_ for clstr in cluster_iterations]
        cluster_iterations = [Leiden_Unimodal(resolution=i) for i in np.linspace(0.1,2.0,10)]
        parts = [clstr.fit_transform(adjacency) for clstr in cluster_iterations]
        modularities = [modularity(adjacency, part) for part in parts]
        # print("Modularity of network clustering: {}".format(np.max(modularities)))
        defacto_network_subgroups = parts[np.argmax(modularities)]
        clusters.append(defacto_network_subgroups)
        ba_matrix = np.zeros((defacto_network_subgroups.size, defacto_network_subgroups.max()+1))
        ba_matrix[np.arange(defacto_network_subgroups.size),defacto_network_subgroups] = 1
        ba_matrices.append(ba_matrix)
        co_matrix = np.matmul(ba_matrix, ba_matrix.T)
        #np.fill_diagonal(co_matrix, 0)
        cluster_graphs.append(csr_matrix(co_matrix))
        # kNN = np.maximum(adjacency, adjacency.T)
        kNNs.append(csr_matrix(adjacency))
elif dataset == 'fftwyt':
    for adjacency in [ff_network, tw_network, yt_network]:
        # cluster_iterations = [Louvain(random_state=int(rand_state), shuffle_nodes=True) for rand_state in np.random.randint(1,high=100000,size=5)]
        # parts = [clstr.fit(adjacency).labels_ for clstr in cluster_iterations]
        cluster_iterations = [Leiden_Unimodal(resolution=i) for i in np.linspace(0.1,2.0,10)]
        parts = [clstr.fit_transform(adjacency) for clstr in cluster_iterations]
        modularities = [modularity(adjacency, part) for part in parts]
        # print("Modularity of network clustering: {}".format(np.max(modularities)))
        defacto_network_subgroups = parts[np.argmax(modularities)]
        clusters.append(defacto_network_subgroups)
        ba_matrix = np.zeros((defacto_network_subgroups.size, defacto_network_subgroups.max()+1))
        ba_matrix[np.arange(defacto_network_subgroups.size),defacto_network_subgroups] = 1
        ba_matrices.append(ba_matrix)
        co_matrix = np.matmul(ba_matrix, ba_matrix.T)
        #np.fill_diagonal(co_matrix, 0)
        cluster_graphs.append(csr_matrix(co_matrix))
        # kNN = np.maximum(adjacency, adjacency.T)
        kNNs.append(csr_matrix(adjacency))
elif dataset == 'aucs':
    for adjacency in [lunch_network, facebook_network, coauthor_network, work_network, leisure_network]:
        # cluster_iterations = [Louvain(random_state=int(rand_state), shuffle_nodes=True) for rand_state in np.random.randint(1,high=100000,size=5)]
        # parts = [clstr.fit(adjacency).labels_ for clstr in cluster_iterations]
        cluster_iterations = [Leiden_Unimodal(resolution=i) for i in np.linspace(0.1,2.0,10)]
        parts = [clstr.fit_transform(adjacency) for clstr in cluster_iterations]
        modularities = [modularity(adjacency, part) for part in parts]
        print("Modularity of network clustering: {}".format(np.max(modularities)))
        defacto_network_subgroups = parts[np.argmax(modularities)]
        clusters.append(defacto_network_subgroups)
        ba_matrix = np.zeros((defacto_network_subgroups.size, defacto_network_subgroups.max()+1))
        ba_matrix[np.arange(defacto_network_subgroups.size),defacto_network_subgroups] = 1
        ba_matrices.append(ba_matrix)
        co_matrix = np.matmul(ba_matrix, ba_matrix.T)
        #np.fill_diagonal(co_matrix, 0)
        cluster_graphs.append(csr_matrix(co_matrix))
        # kNN = np.maximum(adjacency, adjacency.T)
        kNNs.append(csr_matrix(adjacency))
else:
    raise ValueError(dataset+" not found") 


'''
look at the clusters in each mode
'''
if dataset in ['cora', 'citeseer', 'blogcatalog', 'flickr', 'wiki']:
    modes = ['network', 'content']
elif dataset == '3sources':
    modes = ['BBC','Reuters','Guardian']
elif dataset in ['football', 'olympics', 'politicsie', 'politicsuk', 'rugby']:
    modes = ['list content', 'text content', 'follows network','mentions network','retweets network']
elif dataset == 'aucs':
    modes = ['lunch network', 'facebook network', 'coauthor network', 'work network', 'leisure network']
    
# print("Number of ground-truth clusters: {}".format(len(np.unique(y))))

# for idx in range(len(modes)):
#     y_pred = clusters[idx]
#     print("\n")
#     print(modes[idx])
#     print("Adjusted Rand Score: {}".format(adjusted_rand_score(y, y_pred)))
#     print("Adjusted Mutual Information: {}".format(adjusted_mutual_info_score(y, y_pred)))
#     print("Number of Clusters : {}".format(len(np.unique(y_pred))))
#     # print("Confusion Matrix of: \n {}".format(confusion_matrix(y, y_pred)))
    
# for idx_1 in range(len(clusters)-1):
#     for idx_2 in range(idx_1+1,len(clusters)):
#         print("\n")
#         print("Difference between clusters of modes "+modes[idx_1]+" and "+modes[idx_2])
#         print("Adjusted Rand Score: {}".format(adjusted_rand_score(clusters[idx_1], clusters[idx_2])))
#         print("Adjusted Mutual Information: {}".format(adjusted_mutual_info_score(clusters[idx_1], clusters[idx_2])))
#         # print("Confusion Matrix of: \n {}".format(confusion_matrix(clusters[0], clusters[1])))

'''
Look at graphs within each mode
'''
# if dataset in ['cora', 'citeseer', 'blogcatalog', 'flickr', 'wiki']:
#     modes = ['network', 'content']
# elif dataset == '3sources':
#     modes = ['BBC','Reuters','Guardian']
# elif dataset in ['football', 'olympics', 'politicsie', 'politicsuk', 'rugby']:
#     modes = ['list content', 'text content', 'follows network','mentions network','retweets network']

# for idx in range(len(modes)):
#     knn = kNNs[idx]
#     if (abs(knn-knn.T)>1e-10).nnz == 0:
#         G = nx.from_scipy_sparse_matrix(knn)
#     else:
#         G = nx.from_scipy_sparse_matrix(knn, create_using=nx.DiGraph)
#     print("\n")
#     print(modes[idx])
#     print("Density: "+str(nx.density(G)))
#     if nx.is_directed(G):
#         print("Connected Components: "+str(nx.number_weakly_connected_components(G)))
#     else:
#         print("Connected Components: "+str(nx.number_connected_components(G)))
#     print("Average Clustering Coeff: "+str(nx.average_clustering(G)))
#     print("Number of isolates: "+str(nx.number_of_isolates(G)))
#     print("Degree Assortiativity: "+str(nx.degree_assortativity_coefficient(G)))

# for idx_1 in range(len(kNNs)):
#     for idx_2 in range(len(kNNs)):
#         if idx_1 != idx_2:
#             print("Orthogonality between modes "+modes[idx_1]+" and "+modes[idx_2])
#             print(kNNs[idx_1].multiply(kNNs[idx_2]).count_nonzero()/kNNs[idx_1].count_nonzero())


'''
Cross Network Enhancement
'''
# y_preds =[]
# for iteration in range(num_iterations):
#     CE = gl.cross_enhancement_graph(kernel_graphs=kNNs, tau=0)
#     cne_graph = CE.learn_graph(cluster_graphs)
    
#     ''' cluster via louvain '''
#     cluster_iterations = [Louvain(random_state=int(rand_state), shuffle_nodes=True) for rand_state in np.random.randint(1,high=100000,size=5)]
#     parts = [clstr.fit(cne_graph).labels_ for clstr in cluster_iterations]
#     modularities = [modularity(cne_graph, part) for part in parts]
#     y_preds.append(parts[np.argmax(modularities)])
#     ''' cluster via spectral '''
#     #clstr = SpectralClustering(n_clusters = len(np.unique(y)), affinity='precomputed', assign_labels='discretize', n_jobs=-1)
#     #y_pred = clstr.fit_predict((cne_graph+cnd_graph.T)/2)
# print_results(y_preds, y, "CNE")

'''
Cross Network Diffusion with and without cluster associations
'''
# from Influence_Clustering import CND

# y_preds =[]
# for iteration in range(num_iterations):
#     cnd = CND(iter_thresh=0, iterations=10, interior_term='identity', secondary_term='identity')
#     y_preds.append(cnd.fit_predict(clusters, kNNs))
# print_results(y_preds, y, "CND")

# y_preds =[]
# for iteration in range(num_iterations):
#     cnd_c1 = CND(iter_thresh=0, iterations=10, interior_term='clusters', secondary_term='identity')
#     y_preds.append(cnd_c1.fit_predict(clusters, kNNs))
# print_results(y_preds, y, "CND_C1")

# y_preds =[]
# for iteration in range(num_iterations):
#     cnd_c2 = CND(iter_thresh=0, iterations=10, interior_term='identity', secondary_term='clusters')
#     y_preds.append(cnd_c2.fit_predict(clusters, kNNs))
# print_results(y_preds, y, "CND_C2")

'''
MSIMC
'''
# from Influence_Clustering import MSIMC

# y_preds =[]
# for iteration in range(num_iterations):
#     msimc = MSIMC(lambda_power=1, ro=0, n_clusters=len(np.unique(y)))
#     y_preds.append(msimc.fit_predict(kNNs))
#     # y_preds.append(msimc.fit_predict([csr_matrix(adjacency), csr_matrix(X)]))
#     # y_preds.append(msimc.fit_predict([csr_matrix(list_content), csr_matrix(text_content)]+
#     #                                   [csr_matrix(follows_network), csr_matrix(mentions_network), csr_matrix(retweets_network)]))
# print_results(y_preds, y, "MSIMC")

'''
Influence Clustering with and without logic constraints
'''
# import Influence_Clustering as IC

# y_preds =[]
# for iteration in range(num_iterations):
#     cmic = IC.CMIC(alpha=0.9, iterations=30)
#     weighted_ba_matrix = cmic.fit_predict(clusters, kNNs)
#     clstr = BiLouvain()
#     clstr.fit(weighted_ba_matrix)
#     parts = clstr.row_labels_
#     secondary_parts = clstr.col_labels_
#     y_preds.append(parts.copy())
    
#     # clstr = SpectralCoclustering(n_clusters=(len(np.unique(y)), 2)).fit(weighted_ba_matrix)
#     # y_pred = clstr.row_labels_
# print_results(y_preds, y, "CMIC")

# y_preds =[]
# for iteration in range(num_iterations):
#     cmiwcc = IC.CMIWCC(alpha=0.9, iterations=30)
#     weighted_ba_matrix = cmiwcc.fit_predict(clusters, kNNs)
#     clstr = BiLouvain()
#     clstr.fit(weighted_ba_matrix)
#     parts = clstr.row_labels_
#     secondary_parts = clstr.col_labels_
#     y_preds.append(parts.copy())
    
#     # clstr = SpectralCoclustering(n_clusters=(len(np.unique(y)), 2)).fit(weighted_ba_matrix)
#     # y_pred = clstr.row_labels_
# print_results(y_preds, y, "CMIWCC")

'''
Direct Integration with Spectral
'''
# from Direct_Integration import Spectral

# y_preds =[]
# for iteration in range(num_iterations):
#     spc_add = Spectral(k=len(np.unique(y)), alpha=0.5, model='additive', 
#                             cluster_association='similarity')
#     y_preds.append(spc_add.fit_predict(clusters, kNNs))
# print_results(y_preds, y, "Direct Integration Spectral Additive")

# y_preds =[]
# for iteration in range(num_iterations):
#     spc_multiply = Spectral(k=len(np.unique(y)), model='multiplicative', 
#                             cluster_association='similarity')
#     y_preds.append(spc_multiply.fit_predict(clusters, kNNs))
# print_results(y_preds, y, "Direct Integration Spectral Multipilicative")

# y_preds =[]
# for iteration in range(num_iterations):
#     spc_regularized = Spectral(k=len(np.unique(y)), alpha = 1.0, model='regularized')
#     y_preds.append(spc_regularized.fit_predict(clusters, kNNs))
# print_results(y_preds, y, "Direct Integration Regularized Spectral")

'''
Cross Learning
'''
# from Direct_Integration import Cross_Learning

# y_preds =[]
# for iteration in range(num_iterations):
#     cl_add = Cross_Learning(alpha=0.5, model='additive', cluster_association='similarity')
#     y_preds.append(cl_add.fit_predict(clusters, kNNs))
# print_results(y_preds, y, "Direct Integration Cross learning Additive")

# y_preds =[]
# for iteration in range(num_iterations):
#     cl_multiply = Cross_Learning(model='multiplicative', cluster_association='similarity')
#     y_preds.append(cl_multiply.fit_predict(clusters, kNNs))
# print_results(y_preds, y, "Direct Integration Cross Learning Multiplicative")

# y_preds =[]
# for iteration in range(num_iterations):
#     cl_multiply = Cross_Learning(model='augmented', cluster_association='similarity')
#     y_preds.append(cl_multiply.fit_predict(clusters, kNNs))
# print_results(y_preds, y, "Direct Integration Cross Learning Augmented")

'''
Multi-layer clustering with direct integration
'''
# from Direct_Integration import Multi_Layer

# y_preds =[]
# for iteration in range(num_iterations):
#     ml_add = Multi_Layer(alpha=0.5, model='additive', cluster_association='similarity')
#     y_preds.append(ml_add.fit_predict(clusters, kNNs))
# print_results(y_preds, y, "Direct Integration Multi-layer Additive")

# y_preds =[]
# for iteration in range(num_iterations):
#     ml_multiply = Multi_Layer(model='multiplicative', cluster_association='similarity')
#     y_preds.append(ml_multiply.fit_predict(clusters, kNNs))
# print_results(y_preds, y, "Direct Integration Multi-layer Multiplicative")

'''
BGPA+
'''
# import Bipartite_Ensembling as BE

# y_preds =[]
# for iteration in range(num_iterations):
#     bgf = BE.BGPA(refine_clusters = False)
#     y_preds.append(bgf.fit_predict(clusters))
# print_results(y_preds, y, "BGPA+")
# # cluster_comparison(y_preds[0], clusters, modes)

'''
LWBG
'''
# import Bipartite_Ensembling as BE

# y_preds =[]
# for iteration in range(num_iterations):
#     lwbg = BE.LWBG()
#     y_preds.append(lwbg.fit_predict(clusters))
# print_results(y_preds, y, "LWBG")

'''
SPSL
'''
# from Spectral_Integration_Clustering import self_paced_spectral_learning

# y_preds =[]
# for iteration in range(num_iterations):
#     self_spectral = self_paced_spectral_learning(n_clusters = len(np.unique(y)))
#     y_preds.append(self_spectral.fit_predict(kNNs))
# print_results(y_preds, y, "SPSL")

'''
Spectral Feature Intergration (ala Huan Liu)
'''
# from Spectral_Integration_Clustering import spectral_feature_clustering

# y_preds =[]
# for iteration in range(num_iterations):
#     sfc = spectral_feature_clustering(n_clusters = len(np.unique(y)))
#     y_preds.append(sfc.fit_predict(kNNs))
# print_results(y_preds, y, "Spectral Feature Integration")

'''
CSPA+
'''
# from CA_Ensembling import CSPA

# y_preds =[]
# for iteration in range(num_iterations):
#     cspa_clstr = CSPA(threshold=0.3)
#     y_preds.append(cspa_clstr.fit_predict(clusters))
# print_results(y_preds, y, "CSPA+")

'''
GP-MGLA
'''
# from CA_Ensembling import GP_MGLA

# y_preds =[]
# for iteration in range(10):
#     gp_clstr = GP_MGLA()
#     #gp_clstr = GP_MGLA(metaclustering_alg = 'spectral', n_clusters = len(np.unique(y)))
#     y_preds.append(gp_clstr.fit_predict(clusters))
# print_results(y_preds, y, "GP-MGLA")

'''
MCLA+
'''
# import Meta_Cluster_Ensembling as MC

# y_preds =[]
# for iteration in range(num_iterations):
#     m_clstr = MC.MCLA(cluster_comparison = 'rsc')
#     y_preds.append(m_clstr.fit_predict(clusters))
# print_results(y_preds, y, "MCLA+")

'''
LWMC
'''
# import Meta_Cluster_Ensembling as MC

# y_preds =[]
# for iteration in range(num_iterations):
#     m_clstr = MC.LWMC()
#     y_preds.append(m_clstr.fit_predict(clusters))
# print_results(y_preds, y, "LWMC")

'''
ResK
'''
# import rescal_kmeans as RC

# y_preds =[]
# for iteration in range(num_iterations):
#     rc_clstr = RC.rescal_kmeans(n_clusters=len(np.unique(y)))
#     y_preds.append(rc_clstr.fit_predict(kNNs))
# print_results(y_preds, y, "ResK")

'''
Multi-layer Leiden
'''
# from Leiden_Clustering import Leiden_Multiplex
# import Bipartite_Ensembling as BE

# y_preds =[]
# for iteration in range(num_iterations):
#     lc_clstr = Leiden_Multiplex()
#     y_preds.append(lc_clstr.fit_transform(kNNs))
# print_results(y_preds, y, "Multi-layer leiden on graphs")
# # cluster_comparison(y_preds[0], clusters, modes)

# y_preds =[]
# for iteration in range(num_iterations):
#     y_preds.append(lc_clstr.fit_transform(cluster_graphs))
# print_results(y_preds, y, "Mulit-layer leiden on CA graphs")

# y_preds =[]
# for iteration in range(num_iterations):
#     y_preds.append(lc_clstr.fit_transform(kNNs+cluster_graphs))
# print_results(y_preds, y, "Mulit-layer leiden on CA and kNN graphs")

# m1=10
# m2=10
# ari_results = []
# ami_results = []
# mod_1_results = []
# mod_2_results =[]
# for i in np.linspace(0.1,1.0,m1):
#     for j in np.linspace(0.1,1.0,m2):
#         y_preds =[]
#         modularities =[]
#         for iteration in range(5):
#             lc_clstr = Leiden_Multiplex(obj_types=['RB_Mod','RB_Mod'], resolutions=[i,j])
#             y_preds.append(lc_clstr.fit_transform(kNNs))
#             modularities.append(lc_clstr.modularities)
#         aris = [adjusted_rand_score(y, y_pred) for y_pred in y_preds]
#         amis = [adjusted_mutual_info_score(y, y_pred) for y_pred in y_preds]
#         ari_results.append(np.mean(aris))
#         ami_results.append(np.mean(amis))
#         mod_1_results.append(np.mean([modularity[0] for modularity in modularities]))
#         mod_2_results.append(np.mean([modularity[1] for modularity in modularities]))
          

# y_preds =[]
# for iteration in range(num_iterations):
#     lc_clstr = Leiden_Multiplex()
#     y_pred = lc_clstr.fit_transform(kNNs)
#     ba_matrix = np.zeros((y_pred.size, y_pred.max()+1))
#     ba_matrix[np.arange(y_pred.size), y_pred] = 1
#     IA = ba_matrix @ ba_matrix.T
    
#     B = np.concatenate(ba_matrices, axis=1)
#     CA = B @ B.T
#     CA[CA < len(ba_matrices)] =0
#     CA[CA > 0] =1
#     # bgf = BE.BGPA(refine_clusters = False)
#     # y_pred = bgf.fit_predict(clusters)
#     # ba_matrix = np.zeros((y_pred.size, y_pred.max()+1))
#     # ba_matrix[np.arange(y_pred.size), y_pred] = 1
#     # CA = ba_matrix @ ba_matrix.T
    
#     final_cluster = Leiden_Unimodal()
#     alpha = 1/(len(modes)-1)
#     y_preds.append(final_cluster.fit_transform((alpha*CA+(1-alpha)*IA)))
#     # final_cluster = Leiden_Multiplex()
#     # y_preds.append(final_cluster.fit_transform([csr_matrix(IA), csr_matrix(CA)]))
# print_results(y_preds, y, "Multi-layer leiden on graphs")


'''
NF-CCE
'''
# import sys
# sys.path.append("NF_CCE")
# from csnmf import CSNMF

# y_preds =[]
# for iteration in range(num_iterations):
#     objCSNMF = CSNMF([kNN.todense() for kNN in kNNs], len(np.unique(y)), alpha = 0.0001, init = 'rnda', displ = False)
#     res_csnmf = objCSNMF.fit()
#     H = np.array(res_csnmf.matrices[0])
#     y_preds.append(np.argmax(H, axis=1))
# print_results(y_preds, y, "NF-CCE")

'''
CG
'''
# gc_graph = pd.read_csv(os.path.join('Matlab','Consistent Graph Learning', dataset+'_CG.csv'), header=None).values

# ''' cluster via louvain '''
# cluster_iterations = [Louvain(random_state=int(rand_state), shuffle_nodes=True) for rand_state in np.random.randint(1,high=100000,size=5)]
# parts = [clstr.fit(gc_graph).labels_ for clstr in cluster_iterations]
# modularities = [modularity(gc_graph, part) for part in parts]
# y_pred = parts[np.argmax(modularities)]

# print("\n")
# print("CG Results:")
# print("Adjusted Rand Score: {}".format(adjusted_rand_score(y, y_pred)))
# print("Adjusted Mutual Information: {}".format(adjusted_mutual_info_score(y, y_pred)))
# print("Confusion Matrix of: \n {}".format(confusion_matrix(y, y_pred)))

'''
ETL-MSC
'''
# etl_msc_graph = pd.read_csv(os.path.join('Matlab','ETLMSC', dataset+'_ETLMSC.csv'), header=None).values

# ''' cluster via louvain '''
# cluster_iterations = [Louvain(random_state=int(rand_state), shuffle_nodes=True) for rand_state in np.random.randint(1,high=100000,size=5)]
# parts = [clstr.fit(etl_msc_graph).labels_ for clstr in cluster_iterations]
# modularities = [modularity(etl_msc_graph, part) for part in parts]
# y_pred = parts[np.argmax(modularities)]

# print("\n")
# print("ETL-MSC Results:")
# print("Adjusted Rand Score: {}".format(adjusted_rand_score(y, y_pred)))
# print("Adjusted Mutual Information: {}".format(adjusted_mutual_info_score(y, y_pred)))
# print("Confusion Matrix of: \n {}".format(confusion_matrix(y, y_pred)))


'''
Plotting over different alpha and epsilon values for the similarity-biased
diffusion model of CSIC.
'''
# from mpl_toolkits.mplot3d import Axes3D
# from matplotlib import cm
# import MCPR

# alphas = np.linspace(0.1,0.9,10)
# epsilons = np.linspace(0.1,0.9,10)
# X,Y = np. meshgrid(alphas, epsilons)

# for iterations in [10, 30, 50, 100, 200, 500]:
#     AMIs =[]
#     ARIs=[]
#     mods = []
#     num_subgroups =[]
#     for alpha in alphas:
#         for epsilon in epsilons:
#             mcpr = MCPR.MCPR(iterations=iterations, alpha=alpha, epsilon=epsilon)
#             weighted_ba_matrix = mcpr.fit_predict(clusters, kNNs)
#             clstr = Louvain()
#             clstr.fit(weighted_ba_matrix)
#             parts = clstr.labels_
#             num_subgroups.append(len(np.unique(parts)))
#             secondary_parts = clstr.secondary_labels_
#             mods.append(modularity(weighted_ba_matrix, parts, secondary_parts))
#             y_pred = parts.copy()
        
#             ARIs.append(adjusted_rand_score(y, y_pred))
#             AMIs.append(adjusted_mutual_info_score(y, y_pred))
    

#     fig = plt.figure()
#     ax = fig.add_subplot(111, projection='3d')
#     ax.plot_surface(X, Y, np.reshape(ARIs,(10,10)), cmap=cm.coolwarm, linewidth=0)
#     plt.xlabel(r'$\epsilon$')
#     plt.ylabel(r'$\alpha$')
#     plt.title('ARI of BlogCatalog Clusters at {} iterations'.format(iterations))
    
#     fig = plt.figure()
#     ax = fig.add_subplot(111, projection='3d')
#     ax.plot_surface(X, Y, np.reshape(AMIs,(10,10)), cmap=cm.coolwarm, linewidth=0)
#     plt.xlabel(r'$\epsilon$')
#     plt.ylabel(r'$\alpha$')
#     plt.title('AMI of BlogCatalog Clusters at {} iterations'.format(iterations))

'''
Plotting different epsilon values for each of the different modes,
for different number of iterations
'''
# from mpl_toolkits.mplot3d import Axes3D
# from matplotlib import cm
# import MCPR

# epsilon_1 = np.linspace(0.1,0.9,10)
# epsilon_2 = np.linspace(0.1,0.9,10)
# alpha = 0.3
# X,Y = np. meshgrid(epsilon_1, epsilon_2)

# for iterations in [5, 10, 30, 50]:
#     AMIs =[]
#     ARIs=[]
#     mods = []
#     num_subgroups =[]
#     for e_1 in epsilon_1:
#         for e_2 in epsilon_2:
#             mcpr = MCPR.MCPR(iterations=iterations, alpha=alpha, epsilon=[e_1,e_2])
#             weighted_ba_matrix = mcpr.fit_predict(clusters, kNNs)
#             clstr = Louvain()
#             clstr.fit(weighted_ba_matrix)
#             parts = clstr.labels_
#             num_subgroups.append(len(np.unique(parts)))
#             secondary_parts = clstr.secondary_labels_
#             mods.append(modularity(weighted_ba_matrix, parts, secondary_parts))
#             y_pred = parts.copy()
        
#             ARIs.append(adjusted_rand_score(y, y_pred))
#             AMIs.append(adjusted_mutual_info_score(y, y_pred))
    

#     fig = plt.figure()
#     ax = fig.add_subplot(111, projection='3d')
#     ax.plot_surface(X, Y, np.reshape(ARIs,(10,10)), cmap=cm.coolwarm, linewidth=0)
#     plt.xlabel(r'$\epsilon_2$')
#     plt.ylabel(r'$\epsilon_1$')
#     plt.title('ARI of {} Clusters at {} iterations'.format(dataset, iterations))
    
#     fig = plt.figure()
#     ax = fig.add_subplot(111, projection='3d')
#     ax.plot_surface(X, Y, np.reshape(AMIs,(10,10)), cmap=cm.coolwarm, linewidth=0)
#     plt.xlabel(r'$\epsilon_2$')
#     plt.ylabel(r'$\epsilon_1$')
#     plt.title('AMI of {} Clusters at {} iterations'.format(dataset, iterations))

'''
Goodness of clusters over different number of iterations
'''
# AMIs =[]
# ARIs=[]
# mods = []
# num_subgroups =[]
# iterations = range(1, 100, 5)
# for iteration_number in iterations:
#     mcpr = MCPR.MCPR(alpha=0.3, epsilon=0.9, iterations=iteration_number)
#     weighted_ba_matrix = mcpr.fit_predict(clusters, kNNs)
#     clstr = Louvain()
#     clstr.fit(weighted_ba_matrix)
#     parts = clstr.labels_
#     num_subgroups.append(len(np.unique(parts)))
#     secondary_parts = clstr.secondary_labels_
#     mods.append(modularity(weighted_ba_matrix, parts, secondary_parts))
#     y_pred = parts.copy()

#     ARIs.append(adjusted_rand_score(y, y_pred))
#     AMIs.append(adjusted_mutual_info_score(y, y_pred))

# plt.figure()
# #plt.title("AMI and ARI versus Number of Diffusion Iterations")
# plt.plot(iterations, AMIs, 'r', label='AMI')
# plt.plot(iterations, ARIs, 'b', label='ARI')
# plt.xlabel('Number of Iterations', fontsize=12)
# plt.ylabel('Goodness of Clustering', fontsize=12)
# plt.legend()
# plt.show()

'''
Investigate modularity as an optimization criteria
'''
# from Leiden_Clustering import Leiden_Multiplex
# from Direct_Integration import Multi_Layer

# num_iterations =20

# y_preds =[]
# aris = []
# amis =[]
# mods =[]
# for iteration in range(num_iterations):
#     lc_clstr = Leiden_Multiplex(n_iterations=iteration)
#     # lc_clstr = Multi_Layer(alpha=0.5, model='additive')
#     # lc_clstr = Multi_Layer(model='multiplicative')
#     y_pred = lc_clstr.fit_transform(kNNs)
#     y_preds.append(y_pred)
#     aris.append(adjusted_rand_score(y, y_pred))
#     amis.append(adjusted_mutual_info_score(y, y_pred))
#     mods.append(lc_clstr.modularities)
# total_mods = [np.sum(mod) for mod in mods]
# y_mods = [modularity(graph, y) for graph in kNNs]
# print("\n")
# print("Multilayer leiden on graphs")
# print("Modal Modularities for this model: {}".format(mods[np.argmax(total_mods)]))
# print("Ground truth modualrities for this model: {}".format(y_mods))
# print("Best Modualrity achieved: {}".format(np.max(total_mods)))
# print("Ground truth modularity on model: {}".format(np.sum(y_mods)))
# print("correlation with ARI:\n{}".format(np.corrcoef(total_mods, aris)[0,1]))

# y_preds =[]
# aris = []
# amis =[]
# mods =[]
# for iteration in range(num_iterations):
#     lc_clstr = Leiden_Multiplex(n_iterations=iteration)
#     # lc_clstr = Multi_Layer(alpha=0.5, model='additive')
#     # lc_clstr = Multi_Layer(model='multiplicative')
#     y_pred = lc_clstr.fit_transform(cluster_graphs)
#     y_preds.append(y_pred)
#     aris.append(adjusted_rand_score(y, y_pred))
#     amis.append(adjusted_mutual_info_score(y, y_pred))
#     mods.append(lc_clstr.modularities)
# total_mods = [np.sum(mod) for mod in mods]
# y_mods = [modularity(graph, y) for graph in cluster_graphs]
# print("\n")
# print("Multilayer leiden on CA graphs")
# print("Modal Modularities for this model: {}".format(mods[np.argmax(total_mods)]))
# print("Ground truth modualrities for this model: {}".format(y_mods))
# print("Best Modualrity achieved: {}".format(np.max(total_mods)))
# print("Ground truth modularity on model: {}".format(np.sum(y_mods)))
# print("correlation with ARI:\n{}".format(np.corrcoef(total_mods, aris)[0,1]))

# y_preds =[]
# aris = []
# amis =[]
# mods =[]
# for iteration in range(num_iterations):
#     lc_clstr = Multi_Layer(alpha=0.5, model='additive', n_iterations = iteration)
#     # lc_clstr = Multi_Layer(model='multiplicative')
#     y_pred = lc_clstr.fit_predict(clusters, kNNs)
#     y_preds.append(y_pred)
#     aris.append(adjusted_rand_score(y, y_pred))
#     amis.append(adjusted_mutual_info_score(y, y_pred))
#     mods.append(lc_clstr.modularities)
# total_mods = [np.sum(mod) for mod in mods]
# y_mods = [modularity(graph, y) for graph in lc_clstr.transformed_graphs]
# print("\n")
# print("Additive direct integration with multi-layer leiden")
# print("Modal Modularities for this model: {}".format(mods[np.argmax(total_mods)]))
# print("Ground truth modualrities for this model: {}".format(y_mods))
# print("Best Modualrity achieved: {}".format(np.max(total_mods)))
# print("Ground truth modularity on model: {}".format(np.sum(y_mods)))
# print("correlation with ARI:\n{}".format(np.corrcoef(total_mods, aris)[0,1]))

# y_preds =[]
# aris = []
# amis =[]
# mods =[]
# for iteration in range(num_iterations):
#     # lc_clstr = Multi_Layer(alpha=0.5, model='additive')
#     lc_clstr = Multi_Layer(model='multiplicative', n_iterations = iteration)
#     y_pred = lc_clstr.fit_predict(clusters, kNNs)
#     y_preds.append(y_pred)
#     aris.append(adjusted_rand_score(y, y_pred))
#     amis.append(adjusted_mutual_info_score(y, y_pred))
#     mods.append(lc_clstr.modularities)
# total_mods = [np.sum(mod) for mod in mods]
# y_mods = [modularity(graph, y) for graph in lc_clstr.transformed_graphs]
# print("\n")
# print("Multiplicative direct integration with multi-layer leiden")
# print("Modal Modularities for this model: {}".format(mods[np.argmax(total_mods)]))
# print("Ground truth modualrities for this model: {}".format(y_mods))
# print("Best Modualrity achieved: {}".format(np.max(total_mods)))
# print("Ground truth modularity on model: {}".format(np.sum(y_mods)))
# print("correlation with ARI:\n{}".format(np.corrcoef(total_mods, aris)[0,1]))


















